from typing import List

import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score

from series_intro_recognizer.tp.interval import Interval


def _fit_k(data: np.ndarray) -> int:
    best_k = 2
    best_silhouette_score = -1
    for k in range(2, min(data.size - 1, 10)):  # You can adjust the range as needed
        kmeans = KMeans(n_clusters=k, random_state=0).fit(data)
        labels = kmeans.labels_
        if len(set(labels)) == 1:
            continue

        score = silhouette_score(data, labels, random_state=0)

        if score > best_silhouette_score:
            best_silhouette_score = score
            best_k = k

    return best_k


def kmeans_clustering(values: List[float]) -> int:
    data = np.array(values).reshape(-1, 1)

    best_k = _fit_k(data)
    kmeans = KMeans(n_clusters=best_k, random_state=0).fit(data)
    labels = kmeans.labels_

    clusters = {i: data[labels == i] for i in range(best_k)}

    largest_cluster = max(clusters, key=lambda x: len(clusters[x]))
    largest_cluster_data = clusters[largest_cluster]

    median_of_largest_cluster = np.median(largest_cluster_data)

    return int(median_of_largest_cluster)


def find_best_offset(offsets: List[Interval]) -> Interval:
    """
    Returns the most likely offsets for an audio file.
    """
    start_offsets = [offset.start for offset in offsets]
    end_offsets = [offset.end for offset in offsets]

    start_median = kmeans_clustering(start_offsets)
    end_median = kmeans_clustering(end_offsets)

    return Interval(start_median, end_median)



class Config:
    """
    Config file for the series opening recognizer.

    Attributes:
    RATE: int
        Audio sample rate.
    MIN_SEGMENT_LENGTH_SEC: int
        Minimum length of the intro in seconds.
    MIN_SEGMENT_LENGTH_BEATS: int
        Minimum length of the intro in beats.
    PRECISION_SECS: float
        Precision of the correlation in seconds.
    PRECISION_BEATS: int
        Precision of the correlation in beats.
    SERIES_WINDOW: int
        Number of sequential audio samples to be matched with each other.
        E.g. 5 means that the first sample will be matched with the next 5 samples.
    OFFSET_SEARCHER__SEQUENTIAL_SECS: int
        Number of sequential 'non-intro' seconds that signal the end of the intro.
        Intro is considered to be over if the number of sequential 'non-intro' beats is greater than this value.
    OFFSET_SEARCHER__SEQUENTIAL_INTERVALS: int
        Number of sequential 'non-intro' beats that signal the end of the intro.
    SAVE_INTERMEDIATE_RESULTS: bool
        Save the correlation results to 'correlations' and 'offsets' folders.
        Make sure to create the folder before running the app.
    """

    def __init__(self,
                 rate: int = 44100,
                 min_segment_length_sec: int = 30,
                 precision_secs: float = .5,
                 series_window: int = 5,
                 offset_searcher__sequential_secs: int = 30,
                 save_intermediate_results: bool = False):
        """
        Initialize the configuration
        :param rate: Audio sample rate.
        :param min_segment_length_sec: Minimum length of the intro in seconds.
        :param precision_secs: Precision of the correlation in seconds.
        :param series_window: Number of sequential audio samples to be matched with each other.
        :param offset_searcher__sequential_secs: Number of sequential 'non-intro' seconds that signal the end of the intro.
        :param save_intermediate_results: Save the correlation results.
        """
        self.RATE = rate

        self.MIN_SEGMENT_LENGTH_SEC = min_segment_length_sec
        self.MIN_SEGMENT_LENGTH_BEATS = int(min_segment_length_sec * rate)

        self.PRECISION_SECS = precision_secs
        self.PRECISION_BEATS = int(precision_secs * rate)

        self.SERIES_WINDOW = series_window

        self.OFFSET_SEARCHER__SEQUENTIAL_SECS = offset_searcher__sequential_secs
        self.OFFSET_SEARCHER__SEQUENTIAL_INTERVALS = int(offset_searcher__sequential_secs / precision_secs)

        self.SAVE_INTERMEDIATE_RESULTS = save_intermediate_results



from series_intro_recognizer.config import Config
from series_intro_recognizer.processors.audio_files import recognise_from_audio_files


def test_recognise_from_audio_files():
    """
    Copy 8 6-minute audio files to assets/audio_files/ directory and run the test.
    :return:
    """
    cfg = Config()
    files = [(f'assets/audio_files/{i}.wav', None, None) for i in range(1, 8)]
    recognised = recognise_from_audio_files(files, cfg)

    for interval in recognised:
        print(interval, interval.end - interval.start)
        assert interval.start >= 0
        assert interval.end >= 0
        assert interval.end - interval.start > 0
        assert 90 - (interval.end - interval.start) <= 1



# -*- coding: utf-8 -*-

from series_intro_recognizer.tp.interval import Interval
from series_intro_recognizer.processors.audio_samples import recognise_from_audio_samples
from series_intro_recognizer.processors.audio_files import recognise_from_audio_files
